/*
 * Author: Tomek Blaszczyk
 * e-mail: tombla@mimuw.edu.pl
 * Project: Linux Scheduler
 * Release: 1.0.0 1999.09.18
 */
#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>

extern int printk(const char* fmt, ...);


/*----SCHEDULING MODULE IMPLEMENTATION GOES HERE-----*/

#define bsd_boost(task)		(task)->ts_data[0]
#define bsd_starttq(task)	(task)->ts_data[1]
#define bsd_tq(task)		(task)->ts_data[2]

int bsd_sched_on(void)
{
	struct task_struct* task;

	for_each_task(task) {
		task->counter = 0;
		bsd_boost(task) = 0;
		bsd_starttq(task) = 0;
		bsd_tq(task) = 0;
	}
	need_resched = 1;
	return 0;
}
		

int bsd_sched_off(void)
{
	struct task_struct* p;

	for_each_task(p) {
		p->counter = 0;
	}
	need_resched = 1;
	return 0;
}


static int bsd_goodness(struct task_struct* task)
{
	int pri;
	
	if (task->policy != SCHED_OTHER)
		return 1000 + task->rt_priority;

	pri = task->priority + bsd_boost(task) - bsd_tq(task);
	
	return ( pri<3*DEF_PRIORITY ? pri : 3*DEF_PRIORITY );
}


struct task_struct* bsd_select(struct task_struct* prev, int cpu)
{
	long c;
	struct task_struct *p, *next;
	
	c = -1000L;
	next = NULL;
	p = init_task.next_run;
	while (p != &init_task) {
		int weight = bsd_goodness(p);
		if (weight > c)
			c = weight, next = p;
		p = p->next_run;
	}
	
	return next;
}


int bsd_wakeup(struct task_struct* task)
{
	bsd_boost(task) += 10;
	
	if ( !need_resched &&
	     !task->next_run &&
	     bsd_goodness(task) > bsd_goodness(current) )
		need_resched = 1;
	
	return TS_NO_MOVE;
}


int bsd_tqexpired(struct task_struct* task)
{
	bsd_tq(task) += bsd_starttq(task) - task->counter;
	if (bsd_tq(task)>task->priority)
		bsd_tq(task) = task->priority;
	task->counter = bsd_starttq(task) = task->priority;
	
	return TS_MOVE_LAST;
}


int bsd_gotcpu(struct task_struct* task)
{
	bsd_boost(task) = 0;
	
	return TS_NO_MOVE;
}


int bsd_lostcpu(struct task_struct* task)
{
	if (task->state == TASK_RUNNING)
		return TS_NO_MOVE;

	bsd_tq(task) += ( bsd_starttq(task) - task->counter );
	if (bsd_tq(task)>task->priority)
		bsd_tq(task) = task->priority;
	bsd_starttq(task) = task->counter;
		
	return TS_NO_MOVE;
}


int bsd_tick(void)
{
#define BOOST_TICK 1
 
	struct task_struct* task;
	static ticks = 0;
	int nr_tasks;
	
	if ( ++ticks < BOOST_TICK ) 
		return 0;
		
	ticks = 0;
	

	/* count active tasks */
	nr_tasks = 0;	
	for_each_task(task) {
		if (task->state == TASK_RUNNING || 
		    task->state == TASK_UNINTERRUPTIBLE ||
		    task->state == TASK_SWAPPING )
			nr_tasks++;
	}

	for_each_task(task) {
		bsd_tq(task) = bsd_tq(task) * 2 * nr_tasks / ( 2 * nr_tasks + 1 );
		if ( bsd_tq(task) <= 0 ) {
			if ( ++bsd_boost(task) > 3*DEF_PRIORITY )
				bsd_boost(task) = 3*DEF_PRIORITY;
		}
		
	}
	
	return 0;
}


int bsd_get_pri ( struct task_struct* task ) {
	return((int) bsd_tq(task));
}

int bsd_get_nice ( struct task_struct* task ) {
	return((int) DEF_PRIORITY - task->priority - bsd_boost(task));
}


/*
 * ---------------- MODULE-SPECIFIC functions -----------------
 */

#ifdef __cplusplus
extern "C" {
#endif


int init_module( void )
/*
 *  int bsd_install(void)
 */
{              
	if ( ts_installed ) {
		printk ( "bsd_install: Alternate scheduler already installed.\n" );
		return -1;
	}
	
	ts_installed = 1;

	ts_sched_on		= bsd_sched_on;
	ts_sched_off		= bsd_sched_off;
	ts_sched_ctl		= NULL;
	ts_select		= bsd_select;
	ts_none_selected	= NULL;
	ts_wakeup		= bsd_wakeup;
	ts_sleepon		= NULL;
	ts_tqexpired		= bsd_tqexpired;
	ts_gotcpu		= bsd_gotcpu;
	ts_lostcpu		= bsd_lostcpu;
	ts_tick			= bsd_tick;
	ts_sched_stat		= NULL;
	ts_get_pri		= bsd_get_pri;
	ts_get_nice		= bsd_get_nice;
	ts_set_nice		= NULL;
	ts_proc_stat		= NULL;
	
	printk ( "bsd2_install: OK.\n" );
		
	return 0;
}


void cleanup_module(void)
/*
 *   void bsd_uninstall(void)
 */
{
	if ( ts_on && ts_sched_off )
		(*ts_sched_off)();
		
	ts_on = 0;	
	
	ts_sched_on		= NULL;
	ts_sched_off		= NULL;
	ts_sched_ctl		= NULL;
	ts_select		= NULL;
	ts_none_selected	= NULL;
	ts_wakeup		= NULL;
	ts_sleepon		= NULL;
	ts_tqexpired		= NULL;
	ts_gotcpu		= NULL;
	ts_lostcpu		= NULL;
	ts_tick			= NULL;
	ts_sched_stat		= NULL;
	ts_get_pri		= NULL;
	ts_get_nice		= NULL;
	ts_set_nice		= NULL;
	ts_proc_stat		= NULL;

	ts_installed = 0;
	
	printk ( "bsd2_uninstall: OK.\n" );
}	

#ifdef __cplusplus
}
#endif
