/*
 * Author: Tomek Blaszczyk
 * e-mail: tombla@mimuw.edu.pl
 * Project: Linux Scheduler
 * Release: 1.0.0 1999.09.18
 */
#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>

extern int printk(const char* fmt, ...);


/*----SCHEDULING MODULE IMPLEMENTATION GOES HERE-----*/

#define sjf_boost(task)		(task)->ts_data[0]
#define sjf_starttq(task)	(task)->ts_data[1]
#define sjf_tq(task)		(task)->ts_data[2]

int sjf_sched_on(void)
{
	struct task_struct* task;

	for_each_task(task) {
//		printk("SJF:[%d] ctr=%ld pri=%ld\n", task->pid, task->counter, task->priority );
		task->counter = 0;
		sjf_boost(task) = 0;
		sjf_starttq(task) = 0;
		sjf_tq(task) = 0;
	}
	need_resched = 1;
	return 0;
}
		

int sjf_sched_off(void)
{
	struct task_struct* p;

	for_each_task(p) {
		p->counter = 0;
	}
	need_resched = 1;
	return 0;
}


static int sjf_goodness(struct task_struct* task)
{
	int pri;
	
	if (task->policy != SCHED_OTHER)
		return 1000 + task->rt_priority;

	pri = task->priority + sjf_boost(task) - sjf_tq(task);
	
	return ( pri<3*DEF_PRIORITY ? pri : 3*DEF_PRIORITY );
}


struct task_struct* sjf_select(struct task_struct* prev, int cpu)
{
	long c;
	struct task_struct *p, *next;
	
	c = -1000L;
	next = NULL;
	p = init_task.next_run;
	while (p != &init_task) {
		int weight = sjf_goodness(p);
		if (weight > c)
			c = weight, next = p;
		p = p->next_run;
	}
	
	return next;
}


int sjf_wakeup(struct task_struct* task)
{
	sjf_boost(task) += 10;

	if ( !need_resched &&
	     !task->next_run &&
	     sjf_goodness(task) > sjf_goodness(current) )
		need_resched = 1;
		
	return TS_NO_MOVE;
}


int sjf_tqexpired(struct task_struct* task)
{
	sjf_tq(task) = ( sjf_tq(task) + sjf_starttq(task) - task->counter ) / 2;
	task->counter = sjf_starttq(task) = task->priority;
	
	return TS_MOVE_LAST;
}


int sjf_gotcpu(struct task_struct* task)
{
	sjf_boost(task) = 0;

	return TS_NO_MOVE;
}


int sjf_lostcpu(struct task_struct* task)
{
	if ( task->state == TASK_RUNNING )
		return TS_NO_MOVE;
		
	sjf_tq(task) = ( sjf_tq(task) + sjf_starttq(task) - task->counter ) / 2;
	sjf_starttq(task) = task->counter;
	
	return TS_NO_MOVE;
}


int sjf_tick(void)
{
#define BOOST_TICK 2
 
	struct task_struct* task;
	static ticks = 0;
	
	if ( ++ticks < BOOST_TICK ) 
		return 0;
		
	ticks = 0;
	
//	for ( task = init_task.next_run; task != &init_task; task = task->next_run ) {
	for_each_task(task) {
		if ( ++sjf_boost(task) > 6*DEF_PRIORITY )
			sjf_boost(task) = 6*DEF_PRIORITY;
		
	}
	
	return 0;
}


int sjf_get_pri ( struct task_struct* task ) {
	return((int) sjf_tq(task));
}

int sjf_get_nice ( struct task_struct* task ) {
	return((int) sjf_boost(task));
}


/*
 * ---------------- MODULE-SPECIFIC functions -----------------
 */

#ifdef __cplusplus
extern "C" {
#endif


int init_module( void )
/*
 *  int sjf_install(void)
 */
{              
	if ( ts_installed ) {
		printk ( "sjf_install: Alternate scheduler already installed.\n" );
		return -1;
	}
	
	ts_installed = 1;

	ts_sched_on		= sjf_sched_on;
	ts_sched_off		= sjf_sched_off;
	ts_sched_ctl		= NULL;
	ts_select		= sjf_select;
	ts_none_selected	= NULL;
	ts_wakeup		= sjf_wakeup;
	ts_sleepon		= NULL;
	ts_tqexpired		= sjf_tqexpired;
	ts_gotcpu		= sjf_gotcpu;
	ts_lostcpu		= sjf_lostcpu;
	ts_tick			= sjf_tick;
	ts_sched_stat		= NULL;
	ts_get_pri		= sjf_get_pri;
	ts_get_nice		= sjf_get_nice;
	ts_set_nice		= NULL;
	ts_proc_stat		= NULL;
	
	printk ( "sjf_install: OK.\n" );
		
	return 0;
}


void cleanup_module(void)
/*
 *   void sjf_uninstall(void)
 */
{
	if ( ts_on && ts_sched_off )
		(*ts_sched_off)();
		
	ts_on = 0;
	
	ts_sched_on		= NULL;
	ts_sched_off		= NULL;
	ts_sched_ctl		= NULL;
	ts_select		= NULL;
	ts_none_selected	= NULL;
	ts_wakeup		= NULL;
	ts_sleepon		= NULL;
	ts_tqexpired		= NULL;
	ts_gotcpu		= NULL;
	ts_lostcpu		= NULL;
	ts_tick			= NULL;
	ts_sched_stat		= NULL;
	ts_get_pri		= NULL;
	ts_get_nice		= NULL;
	ts_set_nice		= NULL;
	ts_proc_stat		= NULL;

	ts_installed = 0;
	
	printk ( "sjf_uninstall: OK.\n" );
}	

#ifdef __cplusplus
}
#endif
