/*
 * Author: Tomek Blaszczyk
 * e-mail: tombla@mimuw.edu.pl
 * Project: Linux Scheduler
 * Release: 1.0.0 1999.09.18
 */
#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>
#include "tb_svr4.h"

extern int printk(const char* fmt, ...);


/*----SCHEDULING MODULE IMPLEMENTATION GOES HERE-----*/

#define SVR4_WEIGHT	 2
#define SVR4_MAX_TS_PRI  DEF_PRIORITY*2*SVR4_WEIGHT + SVR4_NUM_PRIORITIES 

svr4table_type	svr4table,
		svr4table_tmp;

static int minpid = 100;
static int maxpid = 32767;
static unsigned long dumpflags = SVR4_DUMP_TBLERR;

#define SHOULD_DUMP(flag) (dumpflags & (flag))
#define SHOULD_DUMP_EVENT(task) ((dumpflags & SVR4_DUMP_EVENTS) &&\
	(task)->pid >= minpid &&\
	(task)->pid <= maxpid )

#define svr4_flags(task)	(task)->ts_data[0]
#define svr4_dispwait(task)	(task)->ts_data[1]
#define svr4_dynamic_pri(task)	(task)->ts_data[2]
#define svr4_boost_pri(task)	(task)->ts_data[3]

#define SVR4_VALIDATE(p) svr4_do_validate_task((p))


static void svr4_do_validate_task ( struct task_struct* task )
{
	int dpri, boost;
	
	dpri = svr4_dynamic_pri(task);
	if (  dpri < 0  ||  dpri >= SVR4_NUM_PRIORITIES ) {
		printk("SVR4: Process [%d] has dynamic_pri=%d. Fixed.\n",
			task->pid, dpri );
		if ( dpri < 0 )
			dpri = 0;
		else
			dpri = SVR4_NUM_PRIORITIES - 1;
		svr4_dynamic_pri(task) = dpri;
	}
	
	boost = svr4_boost_pri(task);
	if ( boost< -SVR4_MAX_TS_PRI ||
	     boost > SVR4_MAX_TS_PRI  ) {
		if ( boost < 0 )
			boost = -SVR4_MAX_TS_PRI;
		else
			boost = SVR4_MAX_TS_PRI;
		svr4_boost_pri(task) = boost;
	}
}


int svr4_sched_on(void)
{
	struct task_struct* task;

	for_each_task(task) {
		svr4_flags(task) = 0;
		svr4_dispwait(task) = jiffies;
		svr4_dynamic_pri(task) = task->state ? SVR4_NUM_PRIORITIES - 1 : SVR4_NUM_PRIORITIES / 2;
		svr4_boost_pri(task) = 0;
		task->counter = 0;
	}
	need_resched = 1;
	return 0;
}
		

int svr4_sched_off(void)
{
	struct task_struct* p;

	for_each_task(p) {
		p->counter = 0;
	}
	need_resched = 1;
	return 0;
}


int svr4_set_table(void* data)
{
	int i;
	
	memcpy_fromfs(svr4table_tmp,data,sizeof(svr4table_tmp));
		
	for ( i=0; i<SVR4_NUM_PRIORITIES; i++ ) {
		if (	svr4table_tmp[i].maxwait < 50 ||
			svr4table_tmp[i].maxwait > 2000 ||
			svr4table_tmp[i].counter <=0 ||
			svr4table_tmp[i].counter >200 ||
			svr4table_tmp[i].wakeup_dynamic < 0 ||
			svr4table_tmp[i].wakeup_dynamic >= SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].wakeup_boost <= -SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].wakeup_boost >= SVR4_NUM_PRIORITIES  ||
			svr4table_tmp[i].wakeup_boost_chg <= -SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].wakeup_boost_chg >= SVR4_NUM_PRIORITIES  ||
			svr4table_tmp[i].expired_dynamic < 0 ||
			svr4table_tmp[i].expired_dynamic >= SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].expired_boost <= -SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].expired_boost >= SVR4_NUM_PRIORITIES  ||
			svr4table_tmp[i].expired_boost_chg <= -SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].expired_boost_chg >= SVR4_NUM_PRIORITIES  ||
			svr4table_tmp[i].lwait_dynamic < 0 ||
			svr4table_tmp[i].lwait_dynamic >= SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].lwait_boost <= -SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].lwait_boost >= SVR4_NUM_PRIORITIES  ||
			svr4table_tmp[i].lwait_boost_chg <= -SVR4_NUM_PRIORITIES ||
			svr4table_tmp[i].lwait_boost_chg >= SVR4_NUM_PRIORITIES  ) {
			
			if ( SHOULD_DUMP(SVR4_DUMP_TBLERR) ) {
				printk("SVR4: set_table: entry %d bad.\n", i );
			}
			return -1;
		} /* if */
	} /* for */

	memcpy(svr4table,svr4table_tmp,sizeof(svr4table_type));
		
	return 0;
}  /* svr4_set_table */


int svr4_sched_ctl(int cmd, void* data)
{
	switch ( cmd ) {
	case SVR4_SET_MINPID:
		minpid = (int) data;
		break;
	case SVR4_SET_MAXPID:
		maxpid = (int) data;
		break;
	case SVR4_SET_DUMP_FLAGS:
		dumpflags = (unsigned long) data;
		break;
	case SVR4_GET_TABLE:
		memcpy_tofs(data,svr4table,sizeof(svr4table_type));
		break;
	case SVR4_SET_TABLE:
		return svr4_set_table(data);
	default:
		return -1;
	}
	
	return 0;
}


static long svr4_goodness(struct task_struct* task, struct task_struct* prev)
{
	int pri;

	if (task==&init_task)
		return -1000;
	
	if (task->policy != SCHED_OTHER)
		return 1000 + task->rt_priority;

	pri = SVR4_WEIGHT*task->priority + svr4_dynamic_pri(task) + svr4_boost_pri(task);
	
	if ( task == prev )
		pri += 0;
		
	return ( pri<SVR4_MAX_TS_PRI ? pri : SVR4_MAX_TS_PRI-1 );
}


struct task_struct* svr4_select(struct task_struct* prev, int cpu)
{
	long c;
	struct task_struct *p, *next;
	
	c = -1000L;
	next = NULL;
	p = init_task.next_run;
	while (p != &init_task) {
		int weight = svr4_goodness(p,prev);
		if (weight > c)
			c = weight, next = p;
		p = p->next_run;
	}
	
	return next;
}


int svr4_none_selected(int cpu)
{
	return 0;
}


int svr4_wakeup(struct task_struct* task)
{
	int dpri;
	int boost;

	dpri = svr4_dynamic_pri(task);
	svr4_dynamic_pri(task) = svr4table[dpri].wakeup_dynamic;
	
	boost = svr4table[dpri].wakeup_boost;
	if ( boost )
		svr4_boost_pri(task) = boost;
	else
		svr4_boost_pri(task) += svr4table[dpri].wakeup_boost_chg;

	SVR4_VALIDATE(task);
	
	if ( SHOULD_DUMP_EVENT(task) ) {
		printk("SVR4: [%d] wakeup, dyna=%ld, boost=%ld\n",
			task->pid,
			svr4_dynamic_pri(task),
			svr4_boost_pri(task)       );
	}
	
	if ( !need_resched
	     && !task->next_run 
	     && svr4_goodness(task,current) >= svr4_goodness(current,current)
	   )
		need_resched = 1;
		
	return TS_NO_MOVE;		
}	


/* ts_sleepon - not implemented */


int svr4_tqexpired(struct task_struct* task)
{
	int dpri, boost;

	dpri = svr4_dynamic_pri(task);
	svr4_dynamic_pri(task) = svr4table[dpri].expired_dynamic;
	
	boost = svr4table[dpri].expired_boost;
	if ( boost )
		svr4_boost_pri(task) = boost;
	else
		svr4_boost_pri(task) += svr4table[dpri].expired_boost_chg;
		
	SVR4_VALIDATE(task);
		
	dpri = svr4_dynamic_pri(task);
	task->counter = svr4table[dpri].counter;
	svr4_dispwait(task) = jiffies;

	if ( SHOULD_DUMP_EVENT(task) ) {
		printk ( "SVR4: [%d] expired. dyna=%ld, boost=%ld, quantum=%ld\n",
			task->pid,
			svr4_dynamic_pri(task),
			svr4_boost_pri(task),
			task->counter );
	}
	
	return TS_MOVE_LAST;
}


int svr4_gotcpu(struct task_struct* task)
{
	svr4_boost_pri(task) = 0;

	return TS_NO_MOVE;
}


int svr4_tick(void)
{
	struct task_struct* task;
	
//	for ( task = init_task.next_run; task != &init_task; task = task->next_run ) {
	for_each_task(task) {
	
		int dpri, boost;
		
		dpri = svr4_dynamic_pri(task);
		
		if ( (jiffies-svr4_dispwait(task)) > svr4table[dpri].maxwait ) {
		
			svr4_dispwait(task) = jiffies;
			svr4_dynamic_pri(task) = svr4table[dpri].lwait_dynamic;
			
			boost = svr4table[dpri].lwait_boost;
			if ( boost )
				svr4_boost_pri(task) = boost;
			else
				svr4_boost_pri(task) += svr4table[dpri].lwait_boost_chg;
				
			SVR4_VALIDATE(task);
			
			if ( SHOULD_DUMP_EVENT(task) ) {
				printk("SVR4: [%d] lwait, dyna=%ld, boost=%ld\n",
					task->pid,
					svr4_dynamic_pri(task),
					svr4_boost_pri(task)       );
			}
		}
	}
	
	return 0;
}


int svr4_fork(struct task_struct* parent, struct task_struct* child)
{
	svr4_dispwait(child)=jiffies;
	svr4_boost_pri(child) = 0;
	svr4_dynamic_pri(child) = SVR4_NUM_PRIORITIES / 2;
	
	svr4_boost_pri(parent) = 0;	
	svr4_dynamic_pri(parent) = SVR4_NUM_PRIORITIES / 2;
	
	return 0;
}

/*
 * ts_sched_stat not implemented
 */
int svr4_get_pri ( struct task_struct* task ) {
	return((int) svr4_dynamic_pri(task));
}
int svr4_get_nice ( struct task_struct* task ) {
	return((int) svr4_boost_pri(task));
}
/*
 * ts_set_nice not implemented
 */
/*
 * ts_proc_stat not implemented
 */


/*
 * ---------------- MODULE-SPECIFIC functions -----------------
 */

#ifdef __cplusplus
extern "C" {
#endif

int normalize(int k)
{
	if (k<0) return 0;
	if (k>SVR4_NUM_PRIORITIES-1) return(SVR4_NUM_PRIORITIES-1);
	return k;
}

int init_module( void )
/*
 *  int svr4_install(void)
 */
{              
	int pri;

	if ( ts_installed ) {
		printk ( "svr4_install: Alternate scheduler already installed.\n" );
		return -1;
	}
	
	ts_installed = 1;

	ts_sched_on		= svr4_sched_on;
	ts_sched_off		= svr4_sched_off;
	ts_sched_ctl		= svr4_sched_ctl;
	ts_select		= svr4_select;
	ts_none_selected	= svr4_none_selected;
	ts_wakeup		= svr4_wakeup;
	ts_sleepon		= NULL;
	ts_tqexpired		= svr4_tqexpired;
	ts_gotcpu		= svr4_gotcpu;
	ts_lostcpu		= NULL;
	ts_tick			= svr4_tick;
	ts_fork			= svr4_fork;
	ts_sched_stat		= NULL;
	ts_get_pri		= svr4_get_pri;
	ts_get_nice		= svr4_get_nice;
	ts_set_nice		= NULL;
	ts_proc_stat		= NULL;
	
	for ( pri=0; pri<SVR4_NUM_PRIORITIES; pri++ )
	{
		//
		// This values are taken from 
		//
		
		svr4table[pri].maxwait = 300;
		svr4table[pri].counter = 10+(SVR4_NUM_PRIORITIES-pri)/3;
		svr4table[pri].wakeup_dynamic = normalize(pri+1);
		svr4table[pri].wakeup_boost = 0;
		svr4table[pri].wakeup_boost_chg = 9;
		svr4table[pri].expired_dynamic = normalize(pri-7);
		svr4table[pri].expired_boost = 0; 
		svr4table[pri].expired_boost_chg = 0;
		svr4table[pri].lwait_dynamic = normalize(pri+5);
		svr4table[pri].lwait_boost =  0;
		svr4table[pri].lwait_boost_chg = 6;
	}
	
	printk ( "svr4_install: OK.\n" );
		
	return 0;
}

void cleanup_module(void)
/*
 *   void svr4_uninstall(void)
 */
{
	if ( ts_on && ts_sched_off )
		(*ts_sched_off)();
		
	ts_on = 0;	
	
	ts_sched_on		= NULL;
	ts_sched_off		= NULL;
	ts_sched_ctl		= NULL;
	ts_select		= NULL;
	ts_none_selected	= NULL;
	ts_wakeup		= NULL;
	ts_sleepon		= NULL;
	ts_tqexpired		= NULL;
	ts_gotcpu		= NULL;
	ts_lostcpu		= NULL;
	ts_tick			= NULL;
	ts_fork			= NULL;
	ts_sched_stat		= NULL;
	ts_get_pri		= NULL;
	ts_get_nice		= NULL;
	ts_set_nice		= NULL;
	ts_proc_stat		= NULL;

	ts_installed = 0;
	
	printk ( "nt_uninstall: OK.\n" );
}	

#ifdef __cplusplus
}
#endif
