/*
 * Author: Tomek Blaszczyk
 * e-mail: tombla@mimuw.edu.pl
 * Project: Linux Scheduler
 * Release: 1.0.0 1999.09.18
 */
#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>

extern int printk(const char* fmt, ...);


/*----SCHEDULING MODULE IMPLEMENTATION GOES HERE-----*/

#define unix_boost(task)		(task)->ts_data[0]
#define unix_starttq(task)	(task)->ts_data[1]
#define unix_tq(task)		(task)->ts_data[2]

int unix_sched_on(void)
{
	struct task_struct* task;

	for_each_task(task) {
//		printk("SJF:[%d] ctr=%ld pri=%ld\n", task->pid, task->counter, task->priority );
		task->counter = 0;
		unix_boost(task) = 0;
		unix_starttq(task) = 0;
		unix_tq(task) = 0;
	}
	need_resched = 1;
	return 0;
}
		

int unix_sched_off(void)
{
	struct task_struct* p;

	for_each_task(p) {
		p->counter = 0;
	}
	need_resched = 1;
	return 0;
}


static int unix_goodness(struct task_struct* task)
{
	int pri;
	
	if (task->policy != SCHED_OTHER)
		return 1000 + task->rt_priority;

	pri = task->priority + unix_boost(task) - unix_tq(task);
	
//	if ( task == prev )
//		pri += 1;
		
	return ( pri<3*DEF_PRIORITY ? pri : 3*DEF_PRIORITY );
}


struct task_struct* unix_select(struct task_struct* prev, int cpu)
{
	long c;
	struct task_struct *p, *next;
	
	c = -1000L;
	next = NULL;
	p = init_task.next_run;
	while (p != &init_task) {
		int weight = unix_goodness(p);
		if (weight > c)
			c = weight, next = p;
		p = p->next_run;
	}
	
	return next;
}


int unix_wakeup(struct task_struct* task)
{
	unix_boost(task) += 10;

	if ( !need_resched &&
	     !task->next_run &&
	     unix_goodness(task) > unix_goodness(current) )
		need_resched = 1;
	
	return TS_NO_MOVE;
}


int unix_tqexpired(struct task_struct* task)
{
	unix_tq(task) += unix_starttq(task) - task->counter;
	if (unix_tq(task)>task->priority)
		unix_tq(task) = task->priority;
	task->counter = unix_starttq(task) = task->priority;
	
	return TS_MOVE_LAST;
}


int unix_gotcpu(struct task_struct* task)
{
	unix_boost(task) = 0;
	
	return TS_NO_MOVE;
}


int unix_lostcpu(struct task_struct* task)
{
	if (task->state == TASK_RUNNING)
		return TS_NO_MOVE;

	unix_tq(task) += ( unix_starttq(task) - task->counter );
	if (unix_tq(task)>task->priority)
		unix_tq(task) = task->priority;
	unix_starttq(task) = task->counter;
		
	return TS_NO_MOVE;
}


int unix_tick(void)
{
#define BOOST_TICK 1
 
	struct task_struct* task;
	static ticks = 0;
	
	if ( ++ticks < BOOST_TICK ) 
		return 0;
		
	ticks = 0;
	
//	for ( task = init_task.next_run; task != &init_task; task = task->next_run ) {
	for_each_task(task) {
		unix_tq(task) /= 2;
		if ( unix_tq(task) <= 0 ) {
			if ( ++unix_boost(task) > 3*DEF_PRIORITY )
				unix_boost(task) = 3*DEF_PRIORITY;
		}
		
	}
	
	return 0;
}


int unix_get_pri ( struct task_struct* task ) {
	return((int) unix_tq(task));
}

int unix_get_nice ( struct task_struct* task ) {
	return((int) DEF_PRIORITY - task->priority - unix_boost(task));
}


/*
 * ---------------- MODULE-SPECIFIC functions -----------------
 */

#ifdef __cplusplus
extern "C" {
#endif


int init_module( void )
/*
 *  int unix_install(void)
 */
{              
	if ( ts_installed ) {
		printk ( "unix_install: Alternate scheduler already installed.\n" );
		return -1;
	}
	
	ts_installed = 1;

	ts_sched_on		= unix_sched_on;
	ts_sched_off		= unix_sched_off;
	ts_sched_ctl		= NULL;
	ts_select		= unix_select;
	ts_none_selected	= NULL;
	ts_wakeup		= unix_wakeup;
	ts_sleepon		= NULL;
	ts_tqexpired		= unix_tqexpired;
	ts_gotcpu		= unix_gotcpu;
	ts_lostcpu		= unix_lostcpu;
	ts_tick			= unix_tick;
	ts_sched_stat		= NULL;
	ts_get_pri		= unix_get_pri;
	ts_get_nice		= unix_get_nice;
	ts_set_nice		= NULL;
	ts_proc_stat		= NULL;
	
	printk ( "unix_install: OK.\n" );
		
	return 0;
}


void cleanup_module(void)
/*
 *   void unix_uninstall(void)
 */
{
	if ( ts_on && ts_sched_off )
		(*ts_sched_off)();
		
	ts_on = 0;	
	
	ts_sched_on		= NULL;
	ts_sched_off		= NULL;
	ts_sched_ctl		= NULL;
	ts_select		= NULL;
	ts_none_selected	= NULL;
	ts_wakeup		= NULL;
	ts_sleepon		= NULL;
	ts_tqexpired		= NULL;
	ts_gotcpu		= NULL;
	ts_lostcpu		= NULL;
	ts_tick			= NULL;
	ts_sched_stat		= NULL;
	ts_get_pri		= NULL;
	ts_get_nice		= NULL;
	ts_set_nice		= NULL;
	ts_proc_stat		= NULL;

	ts_installed = 0;
	
	printk ( "unix_uninstall: OK.\n" );
}	

#ifdef __cplusplus
}
#endif
