/*
 * Author: Tomek Blaszczyk
 * e-mail: tombla@mimuw.edu.pl
 * Project: Linux Scheduler
 * Release: 1.0.0 1999.09.18
 */
#include <stdio.h>
#include <qtimer.h>
#include <qpainter.h>
#include <qlist.h>

#include "tbschshow.hpp"
#include "tbschsrc.hpp"
#include "tbschevt.hpp"

TBSchedShow::TBSchedShow ( QList<TBSchedSource>& srcList, QWidget* parent, const char* name)
	: QWidget ( parent, name )
{
	_sourceList = srcList;
	_sourceList.setAutoDelete ( FALSE );
	_jiffies = 0;
	_period  = 100;
	_running = true;

	_timer = new QTimer ( this );
	
	connect ( _timer, SIGNAL(timeout()), SLOT(timeout()) );
	_timer->start ( _period );
}

void TBSchedShow::timeout()
{
	_jiffies += ( _period / 10 );

	if ( !_running )
		return;

	TBSchedSource* src;
	for ( src = _sourceList.first(); src != NULL; src = _sourceList.next() ) {
		src->setCurrentTime ( _jiffies );
		src->FetchData();
	}
		
	repaint();
}

void TBSchedShow::run()
{
	_running = TRUE;
}

void TBSchedShow::stop()
{
	_running = FALSE;
}

void TBSchedShow::period(unsigned long newPeriod)
{
	_period = newPeriod;
	_timer->changeInterval ( _period );
}

void TBSchedShow::slower()
{
	if ( _period <= 1900 )
		period ( _period + 100u );
}

void TBSchedShow::faster()
{
	if ( _period>=200 )
		period ( _period - 100ul );
	else
		period ( 100ul );
}

unsigned long get_currTime() {
	FILE* f;

	f = fopen("/proc/uptime","r");
	if ( f == NULL )
		return 0;
	
	float r;	
	int res = fscanf ( f, "%f", &r );
	fclose(f);
	
	if ( res != 1 )
		return 0;
		
	return (unsigned long) (r*100.0);
}
	
void TBSchedShow::paintEvent ( QPaintEvent* paintEvent )
{
	QPainter p;
	
	const int height = 30;
	const int border = 5;
	const unsigned long interval = 500;

//	unsigned long currTime = _jiffies;
	unsigned long currTime = get_currTime();

	if ( currTime == 0 ){
		printf ( " TBSchedShow::painEvent: currTime==0 !\n" );
		return;
	}

	if ( !_sourceList.last() )
		return;
		
	int count  = _sourceList.at() + 1;
	
	p.begin(this);
	p.setWindow(-30,0,30+interval,(count+1)*border + count*height);
	
	TBSchedSource* src;
	for ( src = _sourceList.last(); src != NULL; src = _sourceList.prev() )
	{
		//
		int i = _sourceList.at();
	
		//
		// paint the text
		//
		
		// coordinates
		int x = -29;
		int y = i * border + (i+1) * height;
		
		// color
		if ( src->fetched() )
			p.setPen ( black );
		else
			p.setPen ( red );
			
		// text
		QString s;
		s.sprintf ( "%d", src->pid() );
		
		// paint
		p.drawText(x,y,s);
		

		//
		// Looping over events for each task		
		//
		
		// don't get dirty data
		if ( !src->fetched() )
			continue;
		
		// set upper and lower bound we iterate on
		if ( currTime<interval)
			src->setStartTime(0);
		else
			src->setStartTime(currTime-interval);
		src->setEndTime(currTime);


		// loop over events
		TBSchedEvent e;
		// y - coord doesn't change
		y = (i+1)*border+i*height;
		for ( e = src->FirstEvent(); !e.eof; e = src->PrevEvent() ) 
		{
			// color
			QColor c;
			switch ( e.event )
			{
			  case SCHED_GOTCPU:
			  	c = green; break;
			  case SCHED_INRUNQ:
			  	c = red; break;
			  case SCHED_OUTRUNQ:
			  	c = blue; break;
			  default:
			  	c = black;
			}
			
			p.setPen(c);
			p.setBrush(c);
			
			// coordinates
			x = currTime - e.end;
			// y coord doesn't change
		
			// draw!	
			p.drawRect(x,y,e.end-e.start+1,height);
						
		} // next event
		
	} // next task
			
	p.end(); // finish painting				
}
